from typing import Literal
from typing import Optional

from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from fractal_tasks_core.channels import ChannelInputModel
from fractal_tasks_core.channels import OmeroChannel


class InitArgsRegistration(BaseModel):
    """
    Registration init args.

    Passed from `image_based_registration_hcs_init` to
    `calculate_registration_image_based`.

    Attributes:
        reference_zarr_url: zarr_url for the reference image
    """

    reference_zarr_url: str


class InitArgsRegistrationConsensus(BaseModel):
    """
    Registration consensus init args.

    Provides the list of zarr_urls for all acquisitions for a given well

    Attributes:
        zarr_url_list: List of zarr_urls for all the OME-Zarr images in the
            well.
    """

    zarr_url_list: list[str]


class InitArgsCellVoyager(BaseModel):
    """
    Arguments to be passed from cellvoyager converter init to compute

    Attributes:
        image_dir: Directory where the raw images are found
        plate_prefix: part of the image filename needed for finding the
            right subset of image files
        well_ID: part of the image filename needed for finding the
            right subset of image files
        image_extension: part of the image filename needed for finding the
            right subset of image files
        image_glob_patterns: Additional glob patterns to filter the available
            images with
        acquisition: Acquisition metadata needed for multiplexing
    """

    image_dir: str
    plate_prefix: str
    well_ID: str
    image_extension: str
    image_glob_patterns: Optional[list[str]]
    acquisition: Optional[int]


class InitArgsIllumination(BaseModel):
    """
    Dummy model description.

    Attributes:
        raw_path: dummy attribute description.
        subsets: dummy attribute description.
    """

    raw_path: str
    subsets: dict[Literal["C_index"], int] = Field(default_factory=dict)


class InitArgsMIP(BaseModel):
    """
    Init Args for MIP task.

    Attributes:
        origin_url: Path to the zarr_url with the 3D data
    """

    origin_url: str


class MultiplexingAcquisition(BaseModel):
    """
    Input class for Multiplexing Cellvoyager converter

    Attributes:
        image_dir: Path to the folder that contains the Cellvoyager image
            files for that acquisition and the MeasurementData &
            MeasurementDetail metadata files.
        allowed_channels: A list of `OmeroChannel` objects, where each channel
            must include the `wavelength_id` attribute and where the
            `wavelength_id` values must be unique across the list.
    """

    image_dir: str
    allowed_channels: list[OmeroChannel]


class NapariWorkflowsOutput(BaseModel):
    """
    A value of the `output_specs` argument in `napari_workflows_wrapper`.

    Attributes:
        type: Output type (either `label` or `dataframe`).
        label_name: Label name (for label outputs, it is used as the name of
            the label; for dataframe outputs, it is used to fill the
            `region["path"]` field).
        table_name: Table name (for dataframe outputs only).
    """

    type: Literal["label", "dataframe"]
    label_name: str
    table_name: Optional[str] = None

    @validator("table_name", always=True)
    def table_name_only_for_dataframe_type(cls, v, values):
        """
        Check that table_name is set only for dataframe outputs.
        """
        _type = values.get("type")
        if (_type == "dataframe" and (not v)) or (_type != "dataframe" and v):
            raise ValueError(
                f"Output item has type={_type} but table_name={v}."
            )
        return v


class NapariWorkflowsInput(BaseModel):
    """
    A value of the `input_specs` argument in `napari_workflows_wrapper`.

    Attributes:
        type: Input type (either `image` or `label`).
        label_name: Label name (for label inputs only).
        channel: `ChannelInputModel` object (for image inputs only).
    """

    type: Literal["image", "label"]
    label_name: Optional[str]
    channel: Optional[ChannelInputModel]

    @validator("label_name", always=True)
    def label_name_is_present(cls, v, values):
        """
        Check that label inputs have `label_name` set.
        """
        _type = values.get("type")
        if _type == "label" and not v:
            raise ValueError(
                f"Input item has type={_type} but label_name={v}."
            )
        return v

    @validator("channel", always=True)
    def channel_is_present(cls, v, values):
        """
        Check that image inputs have `channel` set.
        """
        _type = values.get("type")
        if _type == "image" and not v:
            raise ValueError(f"Input item has type={_type} but channel={v}.")
        return v



# Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
# University of Zurich
#
# Original authors:
# Tommaso Comparin <tommaso.comparin@exact-lab.it>
# Marco Franzon <marco.franzon@exact-lab.it>
#
# This file is part of Fractal and was originally developed by eXact lab S.r.l.
# <exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
# Institute for Biomedical Research and Pelkmans Lab from the University of
# Zurich.
"""
Task for 3D->2D maximum-intensity projection.
"""
import logging
from typing import Any

import anndata as ad
import dask.array as da
import zarr
from pydantic.decorator import validate_arguments
from zarr.errors import ContainsArrayError

from fractal_tasks_core.ngff import load_NgffImageMeta
from fractal_tasks_core.pyramids import build_pyramid
from fractal_tasks_core.roi import (
    convert_ROIs_from_3D_to_2D,
)
from fractal_tasks_core.tables import write_table
from fractal_tasks_core.tables.v1 import get_tables_list_v1
from fractal_tasks_core.tasks.io_models import InitArgsMIP
from fractal_tasks_core.zarr_utils import OverwriteNotAllowedError


logger = logging.getLogger(__name__)


@validate_arguments
def maximum_intensity_projection(
    *,
    # Fractal parameters
    zarr_url: str,
    init_args: InitArgsMIP,
    # Advanced parameters
    overwrite: bool = False,
) -> dict[str, Any]:
    """
    Perform maximum-intensity projection along Z axis.

    Note: this task stores the output in a new zarr file.

    Args:
        zarr_url: Path or url to the individual OME-Zarr image to be processed.
            (standard argument for Fractal tasks, managed by Fractal server).
        init_args: Intialization arguments provided by
            `create_cellvoyager_ome_zarr_init`.
        overwrite: If `True`, overwrite the task output.
    """
    logger.info(f"{init_args.origin_url=}")
    logger.info(f"{zarr_url=}")

    # Read image metadata
    ngff_image = load_NgffImageMeta(init_args.origin_url)
    # Currently not using the validation models due to wavelength_id issue
    # See #681 for discussion
    # new_attrs = ngff_image.dict(exclude_none=True)
    # Current way to get the necessary metadata for MIP
    group = zarr.open_group(init_args.origin_url, mode="r")
    new_attrs = group.attrs.asdict()

    # Create the zarr image with correct
    new_image_group = zarr.group(zarr_url)
    new_image_group.attrs.put(new_attrs)

    # Load 0-th level
    data_czyx = da.from_zarr(init_args.origin_url + "/0")
    num_channels = data_czyx.shape[0]
    chunksize_y = data_czyx.chunksize[-2]
    chunksize_x = data_czyx.chunksize[-1]
    logger.info(f"{num_channels=}")
    logger.info(f"{chunksize_y=}")
    logger.info(f"{chunksize_x=}")

    # Loop over channels
    accumulate_chl = []
    for ind_ch in range(num_channels):
        # Perform MIP for each channel of level 0
        mip_yx = da.stack([da.max(data_czyx[ind_ch], axis=0)], axis=0)
        accumulate_chl.append(mip_yx)
    accumulated_array = da.stack(accumulate_chl, axis=0)

    # Write to disk (triggering execution)
    try:
        accumulated_array.to_zarr(
            f"{zarr_url}/0",
            overwrite=overwrite,
            dimension_separator="/",
            write_empty_chunks=False,
        )
    except ContainsArrayError as e:
        error_msg = (
            f"Cannot write array to zarr group at '{zarr_url}/0', "
            f"with {overwrite=} (original error: {str(e)}).\n"
            "Hint: try setting overwrite=True."
        )
        logger.error(error_msg)
        raise OverwriteNotAllowedError(error_msg)

    # Starting from on-disk highest-resolution data, build and write to disk a
    # pyramid of coarser levels
    build_pyramid(
        zarrurl=zarr_url,
        overwrite=overwrite,
        num_levels=ngff_image.num_levels,
        coarsening_xy=ngff_image.coarsening_xy,
        chunksize=(1, 1, chunksize_y, chunksize_x),
    )

    # Copy over any tables from the original zarr
    # Generate the list of tables:
    tables = get_tables_list_v1(init_args.origin_url)
    roi_tables = get_tables_list_v1(init_args.origin_url, table_type="ROIs")
    non_roi_tables = [table for table in tables if table not in roi_tables]

    for table in roi_tables:
        logger.info(
            f"Reading {table} from "
            f"{init_args.origin_url=}, convert it to 2D, and "
            "write it back to the new zarr file."
        )
        new_ROI_table = ad.read_zarr(f"{init_args.origin_url}/tables/{table}")
        old_ROI_table_attrs = zarr.open_group(
            f"{init_args.origin_url}/tables/{table}"
        ).attrs.asdict()

        # Convert 3D ROIs to 2D
        pxl_sizes_zyx = ngff_image.get_pixel_sizes_zyx(level=0)
        new_ROI_table = convert_ROIs_from_3D_to_2D(
            new_ROI_table, pixel_size_z=pxl_sizes_zyx[0]
        )
        # Write new table
        write_table(
            new_image_group,
            table,
            new_ROI_table,
            table_attrs=old_ROI_table_attrs,
            overwrite=overwrite,
        )

    for table in non_roi_tables:
        logger.info(
            f"Reading {table} from "
            f"{init_args.origin_url=}, and "
            "write it back to the new zarr file."
        )
        new_non_ROI_table = ad.read_zarr(
            f"{init_args.origin_url}/tables/{table}"
        )
        old_non_ROI_table_attrs = zarr.open_group(
            f"{init_args.origin_url}/tables/{table}"
        ).attrs.asdict()

        # Write new table
        write_table(
            new_image_group,
            table,
            new_non_ROI_table,
            table_attrs=old_non_ROI_table_attrs,
            overwrite=overwrite,
        )

    # Generate image_list_updates
    image_list_update_dict = dict(
        image_list_updates=[
            dict(
                zarr_url=zarr_url,
                origin=init_args.origin_url,
                types=dict(is_3D=False),
            )
        ]
    )
    return image_list_update_dict


if __name__ == "__main__":

    from fractal_tasks_core.tasks._utils import run_fractal_task

    run_fractal_task(
        task_function=maximum_intensity_projection,
        logger_name=logger.name,
    )



# Copyright 2024 (C) Friedrich Miescher Institute for Biomedical Research and
# University of Zurich
#
# Original authors:
# Joel Lüthi <joel.luethi@uzh.ch>
#
# This file is part of Fractal and was originally developed by eXact lab S.r.l.
# <exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
# Institute for Biomedical Research and Pelkmans Lab from the University of
# Zurich.
"""
Fractal task list.
"""
from fractal_tasks_core.dev.task_models import CompoundTask
from fractal_tasks_core.dev.task_models import NonParallelTask
from fractal_tasks_core.dev.task_models import ParallelTask

TASK_LIST = [
    CompoundTask(
        name="Convert Cellvoyager to OME-Zarr",
        executable_init="tasks/cellvoyager_to_ome_zarr_init.py",
        executable="tasks/cellvoyager_to_ome_zarr_compute.py",
        meta_init={"cpus_per_task": 1, "mem": 4000},
        meta={"cpus_per_task": 1, "mem": 4000},
    ),
    CompoundTask(
        name="Convert Cellvoyager Multiplexing to OME-Zarr",
        executable_init="tasks/cellvoyager_to_ome_zarr_init_multiplex.py",
        executable="tasks/cellvoyager_to_ome_zarr_compute.py",
        meta_init={"cpus_per_task": 1, "mem": 4000},
        meta={"cpus_per_task": 1, "mem": 4000},
    ),
    CompoundTask(
        name="Maximum Intensity Projection HCS Plate",
        input_types={"is_3D": True},
        executable_init="tasks/copy_ome_zarr_hcs_plate.py",
        executable="tasks/maximum_intensity_projection.py",
        output_types={"is_3D": False},
        meta_init={"cpus_per_task": 1, "mem": 4000},
        meta={"cpus_per_task": 1, "mem": 4000},
    ),
    ParallelTask(
        name="Illumination Correction",
        input_types=dict(illumination_corrected=False),
        executable="tasks/illumination_correction.py",
        output_types=dict(illumination_corrected=True),
        meta={"cpus_per_task": 1, "mem": 4000},
    ),
    # CompoundTask(
    #     name="Illumination Correction (channel parallelized)",
    #     input_types=dict(illumination_correction=False),
    #     executable_init="illumination_correction_init.py",
    #     executable="illumination_correction_compute.py",
    #     output_types=dict(illumination_correction=True),
    # ),
    ParallelTask(
        name="Cellpose Segmentation",
        executable="tasks/cellpose_segmentation.py",
        meta={"cpus_per_task": 4, "mem": 16000, "needs_gpu": True},
    ),
    CompoundTask(
        name="Calculate Registration (image-based)",
        executable_init="tasks/image_based_registration_hcs_init.py",
        executable="tasks/calculate_registration_image_based.py",
        meta_init={"cpus_per_task": 1, "mem": 1000},
        meta={"cpus_per_task": 1, "mem": 8000},
    ),
    CompoundTask(
        name="Find Registration Consensus",
        executable_init="tasks/init_group_by_well_for_multiplexing.py",
        executable="tasks/find_registration_consensus.py",
        meta_init={"cpus_per_task": 1, "mem": 1000},
        meta={"cpus_per_task": 1, "mem": 1000},
    ),
    ParallelTask(
        name="Apply Registration to Image",
        input_types=dict(registered=False),
        executable="tasks/apply_registration_to_image.py",
        output_types=dict(registered=True),
        meta={"cpus_per_task": 1, "mem": 4000},
    ),
    NonParallelTask(
        name="Import OME-Zarr",
        executable="tasks/import_ome_zarr.py",
    ),
    ParallelTask(
        name="Napari Workflows Wrapper",
        executable="tasks/napari_workflows_wrapper.py",
        meta={
            "cpus_per_task": 8,
            "mem": 32000,
        },
    ),
]



# Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
# University of Zurich
#
# Original authors:
# Tommaso Comparin <tommaso.comparin@exact-lab.it>
# Marco Franzon <marco.franzon@exact-lab.it>
#
# This file is part of Fractal and was originally developed by eXact lab S.r.l.
# <exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
# Institute for Biomedical Research and Pelkmans Lab from the University of
# Zurich.
"""
Task that copies the structure of an OME-NGFF zarr array to a new one.
"""
import logging
from typing import Any

import zarr
from pydantic.decorator import validate_arguments

import fractal_tasks_core
from fractal_tasks_core.ngff.specs import NgffPlateMeta
from fractal_tasks_core.ngff.specs import WellInPlate
from fractal_tasks_core.ngff.zarr_utils import load_NgffPlateMeta
from fractal_tasks_core.ngff.zarr_utils import load_NgffWellMeta
from fractal_tasks_core.tasks.io_models import InitArgsMIP
from fractal_tasks_core.zarr_utils import open_zarr_group_with_overwrite

logger = logging.getLogger(__name__)


__OME_NGFF_VERSION__ = fractal_tasks_core.__OME_NGFF_VERSION__


def _get_plate_url_from_image_url(zarr_url: str) -> str:
    """
    Given the absolute `zarr_url` for an OME-Zarr image within an HCS plate,
    return the path to the plate zarr group.
    """
    zarr_url = zarr_url.rstrip("/")
    plate_path = "/".join(zarr_url.split("/")[:-3])
    return plate_path


def _get_well_sub_url(zarr_url: str) -> str:
    """
    Given the absolute `zarr_url` for an OME-Zarr image within an HCS plate,
    return the path to the image zarr group.
    """
    zarr_url = zarr_url.rstrip("/")
    well_url = "/".join(zarr_url.split("/")[-3:-1])
    return well_url


def _get_image_sub_url(zarr_url: str) -> str:
    """
    Given the absolute `zarr_url` for an OME-Zarr image, return the image
    zarr-group name.
    """
    zarr_url = zarr_url.rstrip("/")
    image_sub_url = zarr_url.split("/")[-1]
    return image_sub_url


def _generate_wells_rows_columns(
    well_list: list[str],
) -> tuple[list[WellInPlate], list[str], list[str]]:
    """
    Generate the plate well metadata based on the list of wells.
    """
    rows = []
    columns = []
    wells = []
    for well in well_list:
        rows.append(well.split("/")[0])
        columns.append(well.split("/")[1])
    rows = sorted(list(set(rows)))
    columns = sorted(list(set(columns)))
    for well in well_list:
        wells.append(
            WellInPlate(
                path=well,
                rowIndex=rows.index(well.split("/")[0]),
                columnIndex=columns.index(well.split("/")[1]),
            )
        )

    return wells, rows, columns


def _generate_plate_well_metadata(
    zarr_urls: list[str],
) -> tuple[dict[str, dict], dict[str, dict[str, dict]], dict[str, dict]]:
    """
    Generate metadata for OME-Zarr HCS plates & wells.

    Based on the list of zarr_urls, generate metadata for all plates and all
    their wells.

    Args:
        zarr_urls: List of paths or urls to the individual OME-Zarr image to
            be processed.

    Returns:
        plate_metadata_dicts: Dictionary of plate plate metadata. The structure
            is: {"old_plate_name": NgffPlateMeta (as dict)}.
        new_well_image_attrs: Dictionary of image lists for the new wells.
            The structure is: {"old_plate_name": {"old_well_name":
            [ImageInWell(as dict)]}}
        well_image_attrs: Dictionary of Image attributes of the existing wells.
    """
    # TODO: Simplify this block. Currently complicated, because we need to loop
    # through all potential plates, all their wells & their images to build up
    # the metadata for the plate & well.
    plate_metadata_dicts = {}
    plate_wells = {}
    well_image_attrs = {}
    new_well_image_attrs = {}
    for zarr_url in zarr_urls:
        # Extract plate/well/image parts of `zarr_url`
        old_plate_url = _get_plate_url_from_image_url(zarr_url)
        well_sub_url = _get_well_sub_url(zarr_url)
        curr_img_sub_url = _get_image_sub_url(zarr_url)

        # The first time a plate is found, create its metadata
        if old_plate_url not in plate_metadata_dicts:
            logger.info(f"Reading plate metadata of {old_plate_url=}")
            old_plate_meta = load_NgffPlateMeta(old_plate_url)
            plate_metadata = dict(
                plate=dict(
                    acquisitions=old_plate_meta.plate.acquisitions,
                    field_count=old_plate_meta.plate.field_count,
                    name=old_plate_meta.plate.name,
                    # The new field count could be different from the old
                    # field count
                    version=old_plate_meta.plate.version,
                )
            )
            plate_metadata_dicts[old_plate_url] = plate_metadata
            plate_wells[old_plate_url] = []
            well_image_attrs[old_plate_url] = {}
            new_well_image_attrs[old_plate_url] = {}

        # The first time a plate/well pair is found, create the well metadata
        if well_sub_url not in plate_wells[old_plate_url]:
            plate_wells[old_plate_url].append(well_sub_url)
            old_well_url = f"{old_plate_url}/{well_sub_url}"
            logger.info(f"Reading well metadata of {old_well_url}")
            well_attrs = load_NgffWellMeta(old_well_url)
            well_image_attrs[old_plate_url][well_sub_url] = well_attrs.well
            new_well_image_attrs[old_plate_url][well_sub_url] = []

        # Find images of the current well with name matching the current image
        # TODO: clarify whether this list must always have length 1
        curr_well_image_list = [
            img
            for img in well_image_attrs[old_plate_url][well_sub_url].images
            if img.path == curr_img_sub_url
        ]
        new_well_image_attrs[old_plate_url][
            well_sub_url
        ] += curr_well_image_list

    # Fill in the plate metadata based on all available wells
    for old_plate_url in plate_metadata_dicts:
        well_list, row_list, column_list = _generate_wells_rows_columns(
            plate_wells[old_plate_url]
        )
        plate_metadata_dicts[old_plate_url]["plate"]["columns"] = []
        for column in column_list:
            plate_metadata_dicts[old_plate_url]["plate"]["columns"].append(
                {"name": column}
            )

        plate_metadata_dicts[old_plate_url]["plate"]["rows"] = []
        for row in row_list:
            plate_metadata_dicts[old_plate_url]["plate"]["rows"].append(
                {"name": row}
            )
        plate_metadata_dicts[old_plate_url]["plate"]["wells"] = well_list

        # Validate with NgffPlateMeta model
        plate_metadata_dicts[old_plate_url] = NgffPlateMeta(
            **plate_metadata_dicts[old_plate_url]
        ).dict(exclude_none=True)

    return plate_metadata_dicts, new_well_image_attrs, well_image_attrs


@validate_arguments
def copy_ome_zarr_hcs_plate(
    *,
    # Fractal parameters
    zarr_urls: list[str],
    zarr_dir: str,
    # Advanced parameters
    suffix: str = "mip",
    overwrite: bool = False,
) -> dict[str, Any]:
    """
    Duplicate the OME-Zarr HCS structure for a set of zarr_urls.

    This task only processes the zarr images in the zarr_urls, not all the
    images in the plate. It copies all the  plate & well structure, but none
    of the image metadata or the actual image data:

    - For each plate, create a new OME-Zarr HCS plate with the attributes for
        all the images in zarr_urls
    - For each well (in each plate), create a new zarr subgroup with the
       same attributes as the original one.

    Note: this task makes use of methods from the `Attributes` class, see
    https://zarr.readthedocs.io/en/stable/api/attrs.html.

    Args:
        zarr_urls: List of paths or urls to the individual OME-Zarr image to
            be processed.
            (standard argument for Fractal tasks, managed by Fractal server).
        zarr_dir: path of the directory where the new OME-Zarrs will be
            created.
            (standard argument for Fractal tasks, managed by Fractal server).
        suffix: The suffix that is used to transform `plate.zarr` into
            `plate_suffix.zarr`. Note that `None` is not currently supported.
        overwrite: If `True`, overwrite the task output.

    Returns:
        A parallelization list to be used in a compute task to fill the wells
        with OME-Zarr images.
    """

    # Preliminary check
    if suffix is None or suffix == "":
        raise ValueError(
            "Running copy_ome_zarr_hcs_plate without a suffix would lead to"
            "overwriting of the existing HCS plates."
        )

    parallelization_list = []

    # Generate parallelization list
    for zarr_url in zarr_urls:
        old_plate_url = _get_plate_url_from_image_url(zarr_url)
        well_sub_url = _get_well_sub_url(zarr_url)
        old_plate_name = old_plate_url.split(".zarr")[-2].split("/")[-1]
        new_plate_name = f"{old_plate_name}_{suffix}"
        zarrurl_plate_new = f"{zarr_dir}/{new_plate_name}.zarr"
        curr_img_sub_url = _get_image_sub_url(zarr_url)
        new_zarr_url = f"{zarrurl_plate_new}/{well_sub_url}/{curr_img_sub_url}"
        parallelization_item = dict(
            zarr_url=new_zarr_url,
            init_args=dict(origin_url=zarr_url),
        )
        InitArgsMIP(**parallelization_item["init_args"])
        parallelization_list.append(parallelization_item)

    # Generate the plate metadata & parallelization list
    (
        plate_attrs_dicts,
        new_well_image_attrs,
        well_image_attrs,
    ) = _generate_plate_well_metadata(zarr_urls=zarr_urls)

    # Create the new OME-Zarr HCS plate
    for old_plate_url, plate_attrs in plate_attrs_dicts.items():
        old_plate_name = old_plate_url.split(".zarr")[-2].split("/")[-1]
        new_plate_name = f"{old_plate_name}_{suffix}"
        zarrurl_new = f"{zarr_dir}/{new_plate_name}.zarr"
        logger.info(f"{old_plate_url=}")
        logger.info(f"{zarrurl_new=}")
        new_plate_group = open_zarr_group_with_overwrite(
            zarrurl_new, overwrite=overwrite
        )
        new_plate_group.attrs.put(plate_attrs)

        # Write well groups:
        for well_sub_url in new_well_image_attrs[old_plate_url]:
            new_well_group = zarr.group(f"{zarrurl_new}/{well_sub_url}")
            well_attrs = dict(
                well=dict(
                    images=[
                        img.dict(exclude_none=True)
                        for img in new_well_image_attrs[old_plate_url][
                            well_sub_url
                        ]
                    ],
                    version=well_image_attrs[old_plate_url][
                        well_sub_url
                    ].version,
                )
            )
            new_well_group.attrs.put(well_attrs)

    return dict(parallelization_list=parallelization_list)


if __name__ == "__main__":
    from fractal_tasks_core.tasks._utils import run_fractal_task

    run_fractal_task(
        task_function=copy_ome_zarr_hcs_plate,
        logger_name=logger.name,
    )
